''*****************************
''*  TV Text 40x13 v1.0       *
''*  (C) 2006 Parallax, Inc.  *
''*****************************

CON
  Bsp          =  8                               ' Backspace
  Tab          =  9                               ' Tab (every 8 columns)
  Lf           = 10                               ' Line feed
  Cr           = 13                               ' Carriage return
  Esc          = 27                               ' Escape

  cols         = 40                               ' Number of display columns
  rows         = 30                               ' Number of display rows
  maxColor     =  8                               ' Number of color palettes

  screensize   = cols * rows                      ' Character cells on display
  blank        = $0220                            ' Tile pointer for blank

  tv_count     = 14                               ' Number of parameter longs

  screen       = $8000 - screensize               ' Screen buffer
  colors       = screen - 8 * 8                   ' Color table for 8 colors
  cursor       = colors - 3                       ' Cursor information
                                                  '  +0 - current column
                                                  '  +1 - current row
                                                  '  +2 - current col
  cogUse       = cursor - 1                       ' Cog used by TV driver
  sync         = cogUse - tv_count * 4            ' TV driver parameter table
  
  tv_status    = sync +  0 '0/1/2 = off/invisible/visible            read-only
  tv_enable    = sync +  4 '0/non-0 = off/on                         write-only
  tv_pins      = sync +  8 '%pppmmmm = pin group, pin group mode     write-only
  tv_mode      = sync + 12 '%tccip = tile,chroma,interlace,ntsc/pal  write-only
  tv_screen    = sync + 16 'pointer to screen (words)                write-only
  tv_colors    = sync + 20 'pointer to colors (longs)                write-only
  tv_ht        = sync + 24 'horizontal tiles                         write-only
  tv_vt        = sync + 28 'vertical tiles                           write-only
  tv_hx        = sync + 32 'horizontal tile expansion                write-only
  tv_vx        = sync + 36 'vertical tile expansion                  write-only
  tv_ho        = sync + 40 'horizontal offset                        write-only
  tv_vo        = sync + 44 'vertical offset                          write-only
  tv_broadcast = sync + 48 'broadcast frequency (Hz)                 write-only
  tv_auralcog  = sync + 52 'aural fm cog                             write-only

VAR
  long state, vt100Val[4]
  byte vt100Cnt, vt100Def[4]

OBJ

  tv : "tv_half_height"


PUB start(basepin) : okay

'' Start terminal - starts a cog
'' returns false if no cog available
  long[cursor] := state := 0
  longmove(tv_status, @tv_params, tv_count)
  long[tv_pins]   := (basepin & $38) << 1 | (basepin & 4 == 4) & %0101
  long[tv_screen] := screen
  long[tv_colors] := colors
  str(string(Esc,"[1;007;010u"))                  '  white / dark blue $07/$0A
  str(string(Esc,"[2;007;187u"))                  '  white / red       $07/$BB
  str(string(Esc,"[3;158;155u"))                  ' yellow / brown     $9E/$9B
  str(string(Esc,"[4;004;007u"))                  '   grey / white     $04/$07
  str(string(Esc,"[5;061;059u"))                  '   cyan / dark cyan $3D/$3B
  str(string(Esc,"[6;107;110u"))                  '  green / dark green$6B/$6E
  str(string(Esc,"[7;187;206u"))                  '    red / pink      $BB/$CE
  str(string(Esc,"[8;060;010u"))                  '   cyan / blue      $3C/$0A
  str(string(Esc,"[2K"))                          ' Clear screen
  byte[cogUse] := okay := tv.start(tv_status)

PUB stop

'' Stop terminal - frees a cog

  if byte[cogUse]                                 ' Works the same as tv.stop
    cogstop(byte[cogUse]~ - 1)

PUB str(stringptr)

'' Print a zero-terminated string

  repeat strsize(stringptr)
    out(byte[stringptr++])

PUB dec(value) | i

'' Print a decimal number

  if value < 0
    -value
    out("-")

  i := 1_000_000_000

  repeat 10
    if value => i
      out(value / i + "0")
      value //= i
      result~~
    elseif result or i == 1
      out("0")
    i /= 10

PUB hex(value, digits)

'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    out(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)

'' Print a binary number

  value <<= 32 - digits
  repeat digits
    out((value <-= 1) & 1 + "0")

PUB out(c) | n, f, b                              ' Display a character
' This routine implements a subset of ECMA-48 display control codes.
' Bsp                       - Backspace (destructive)
' Tab                       - Tab (every 8 columns)
' Lf                        - Line feed
' Cr                        - Carriage return
' Esc E                     - New line (like Cr/Lf)
' Esc D                     - Index (like Lf)
' Esc M                     - Reverse index (reverse Lf)
' Esc [ <params> <code>     - CSI (see below)
'                           -  <params> are of the form: 0;1;2;3
'                           -  1 to 4 semicolon delimited digit strings
' CSI H                     - Move cursor to <x,y>.  Defaults <1,1>
' CSI A                     - Cursor up # lines.  Default 1
' CSI B                     - Cursor down # lines.  Default 1
' CSI C                     - Cursor forward (right) # cols.  Default 1
' CSI D                     - Cursor backward (left) # cols.  Default 1
' CSI K                     - Erase characters (default is option 0)
'                           -  0 - from cursor to end of line
'                           -  1 - from beginning of line to cursor
'                           -  2 - entire line containing cursor
' CSI J                     - Erase lines (default is option 0)
'                           -  0 - from current line to bottom of screen
'                           -  1 - from top of screen to current line
'                           -  2 - entire screen.  Also homes cursor
' CSI u                     - Set color to parameter if only one parameter
'                           - Set color to foreground / background values
' CSI v                     - Roll up screen region <col,row,#cols,#rows>
' CSI w                     - Roll down screen region <col,row,#cols,#rows>
' CSI x                     - Clear screen region <col,row,#cols,#rows>
  case state
    0: case c & $7F                               ' State: Home
         Bsp:                                     ' Backspace
           if byte[cursor+0] > 0                  '  Backup one column
             word[screen+byte[cursor+1]*cols+(--byte[cursor+0])] := blank
         Tab:
           repeat
             word[screen+byte[cursor+1]*cols+byte[cursor+0]] := blank
             if ++byte[cursor+0] == cols
               newline
           while byte[cursor+0] & 7
         Lf:                                      ' Line Feed
           if ++byte[cursor+1] == rows            '  Go to next row
             rollUpRegion(0,0,cols,rows)          '  Roll up if necessary
             byte[cursor+1] := rows - 1
         Cr:
           byte[cursor+0] := 0                    ' Carriage Return
         Esc:
           state := 1                             ' Escape character
         other:
           result := (byte[cursor+3] << 1 + c & 1) << 10 + $200 + c & $FE
           word[screen][byte[cursor+1] * cols + byte[cursor+0]] := result
           if byte[cursor+0] == cols
             newline
    1: case c & $7F                               ' State: After ESC found
         "[": state := 2
              vt100Cnt := 0
              vt100Val[vt100Cnt]~                 ' Start first parameter
              vt100Def[vt100Cnt++]~~              '  and use default value
         "E": state := 0                          ' Next line
              newline
         "D": state := 0                          ' Index                         
              if ++byte[cursor+1] == rows         ' Go to next row
                rollUpRegion(0,0,cols,rows)       ' Roll up if necessary
                byte[cursor+1] := rows - 1
         "M": state := 0                          ' Reverse index
              if byte[cursor+1]-- == 255          ' Go to previous row
                rollDnRegion(0,0,cols,rows)       ' Roll down if necessary
                byte[cursor+1] := 0
    2: case c & $7F                               ' State: Parameters after "["
         "0".."9":
              vt100Val[--vt100Cnt] *= 10          ' Get explicit value
              vt100Val[vt100Cnt] += c - "0"
              vt100Def[vt100Cnt++]~
         ";": if vt100Cnt < 4                     ' Reuse last parameter if
                vt100Cnt++                        '  too many already
              vt100Val[--vt100Cnt]~               ' Initialize a new parameter
              vt100Def[vt100Cnt++]~~
         "H": state := 0                          ' Move cursor to position
              byte[cursor+0] := parm(0,1)-1 <# cols-1
              byte[cursor+1] := parm(1,1)-1 <# rows-1
         "A": state := 0                          ' Cursor up
              byte[cursor+1] := (byte[cursor+1] - parm(1,1)) #> 0
         "B": state := 0                          ' Cursor down
              byte[cursor+1] := (byte[cursor+1] + parm(1,1)) <# rows-1
         "C": state := 0                          ' Cursor forward (right)
              byte[cursor+0] := (byte[cursor+0] + parm(0,1)) <# cols-1
         "D": state := 0                          ' Cursor backward (left)
              byte[cursor+0] := (byte[cursor+0] - parm(0,1)) #> 0
         "K": state := 0
              f := byte[cursor+0]
              b := byte[cursor+1]
              case parm(0,0)
                0: clearRegion(f,b,cols-f,1)
                1: clearRegion(0,b,b+1,1)         ' Erase to start of line
                2: clearRegion(0,b,cols,1)        ' Erase current line
         "J": state := 0
              b := byte[cursor+1]
              case parm(0,0)
                0: clearRegion(0,b,cols,rows-b)   ' Erase lines to bottom
                1: clearRegion(0,0,cols,b+1)      ' Erase lines to top
                2: clearRegion(0,0,cols,rows)     ' Erase entire screen
                   byte[cursor+0] := 0            '  and home cursor
                   byte[cursor+1] := 0
         "u": state := 0
              n := parm(0,byte[cursor+3]+1)-1 <# maxColor-1
              if vt100Def[1] <> 0 and vt100Def[2] <> 0
                byte[cursor+3] := n + 1           ' Set current color
              else
                if vt100Def[1] <> 0               ' Default foreground color
                  f := long[colors][n<<1] >> 24   '  is current value
                else
                  f := vt100Val[1] & $FF
                if vt100Def[2] <> 0               ' Default background color
                  b := long[colors][n<<1] & $FF   '  is current value
                else
                  b := vt100Val[2] & $FF
                long[colors][n<<1+0] := f << 24 + b << 16 + f << 8 + b
                long[colors][n<<1+1] := f << 24 + f << 16 + b << 8 + b
         "v": state := 0                          ' Roll up screen region
              rollUpRegion(parm(0,1)-1,parm(1,1)-1,parm(2,1),parm(3,1))
         "w": state := 0                          ' Roll down screen region
              rollDnRegion(parm(0,1)-1,parm(1,1)-1,parm(2,1),parm(3,1))
         "x": state := 0                          ' Clear screen region
              clearRegion(parm(0,1)-1,parm(1,1)-1,parm(2,1),parm(3,1))
  
PRI parm(n,d)                       ' Return parameter value or default
   if vt100Def[n]
      result := d
   else
      result := vt100Val[n]

PUB newline                         ' Start a new line
  if ++byte[cursor+1] == rows       ' Go to next row
    rollUpRegion(0,0,cols,rows)     ' Roll up if necessary
    byte[cursor+1] := rows - 1
  byte[cursor+0] := 0               ' Go to start of next line

PUB rollUpRegion(xx,yy,w,h) | i     ' Roll up a region of the screen
  if h > 1
    repeat i from yy to yy+h-2
      wordmove(screen+(cols*i+xx)<<1,screen+(cols*(i+1)+xx)<<1,w)
  wordfill(screen+(cols*(yy+h-1)+xx)<<1,blank,w)

PUB rollDnRegion(xx,yy,w,h) | i     ' Roll down a region of the screen
  if h > 1
     repeat i from yy+h-2 to yy
       wordmove(screen+(cols*(i+1)+xx)<<1,screen+(cols*i+xx)<<1,w)
  wordfill(screen+(cols*yy+xx)<<1,blank,w)

PUB clearRegion(xx,yy,w,h) | i      ' Clear a region of the screen
  repeat i from yy to yy+h-1
     wordfill(screen+(cols*i+xx)<<1,blank,w)

DAT
tv_params               long    0               'status
                        long    1               'enable
                        long    0               'pins
                        long    %10010          'mode
                        long    0               'screen
                        long    0               'colors
                        long    cols            'hc
                        long    rows            'vc
                        long    4               'hx
                        long    1               'vx
                        long    0               'ho
                        long    0               'vo
                        long    0               'broadcast
                        long    0               'auralcog
